'use strict';

var d3Selection = require('d3-selection');
var d3Transition = require('d3-transition');
var d3Zoom = require('d3-zoom');
var d3Axis = require('d3-axis');
var d3Scale = require('d3-scale');
var d3Shape = require('d3-shape');
var d3Array = require('d3-array');
var d3Horizon = require('d3-horizon');
var Kapsule = require('kapsule');
var accessorFn = require('accessor-fn');
var indexBy = require('index-array-by');
var memo = require('lodash.memoize');
var d3Time = require('d3-time');
var d3TimeFormat = require('d3-time-format');

function _interopDefaultLegacy (e) { return e && typeof e === 'object' && 'default' in e ? e : { 'default': e }; }

var d3Horizon__default = /*#__PURE__*/_interopDefaultLegacy(d3Horizon);
var Kapsule__default = /*#__PURE__*/_interopDefaultLegacy(Kapsule);
var accessorFn__default = /*#__PURE__*/_interopDefaultLegacy(accessorFn);
var indexBy__default = /*#__PURE__*/_interopDefaultLegacy(indexBy);
var memo__default = /*#__PURE__*/_interopDefaultLegacy(memo);

function styleInject(css, ref) {
  if (ref === void 0) ref = {};
  var insertAt = ref.insertAt;

  if (!css || typeof document === 'undefined') {
    return;
  }

  var head = document.head || document.getElementsByTagName('head')[0];
  var style = document.createElement('style');
  style.type = 'text/css';

  if (insertAt === 'top') {
    if (head.firstChild) {
      head.insertBefore(style, head.firstChild);
    } else {
      head.appendChild(style);
    }
  } else {
    head.appendChild(style);
  }

  if (style.styleSheet) {
    style.styleSheet.cssText = css;
  } else {
    style.appendChild(document.createTextNode(css));
  }
}

var css_248z = ".horizon-series {\n  position: relative;\n  border: 0 solid darkgray;\n}\n\n.horizon-series:hover {\n  border-top: 1px solid indianred;\n  border-bottom: 1px solid indianred;\n}\n\n.horizon-series .label {\n  position: absolute;\n  bottom: 0;\n  left: 4px;\n  font-family: sans-serif;\n  text-shadow: 0 1px 0 rgba(255, 255, 255, .5);\n  pointer-events: none;\n}\n\n.horizon-series .chart .horizon-tooltip {\n  z-index: 1000;\n}\n\n.horizon-ruler {\n  position: absolute;\n  background: black;\n  opacity: 0;\n  top: 0;\n  bottom: 0;\n  width: 1px;\n  pointer-events: none;\n  transition: opacity 0.5s;\n  z-index: 100;\n}";
styleInject(css_248z);

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);

    if (enumerableOnly) {
      symbols = symbols.filter(function (sym) {
        return Object.getOwnPropertyDescriptor(object, sym).enumerable;
      });
    }

    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(Object(source), true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(Object(source)).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

function _typeof(obj) {
  "@babel/helpers - typeof";

  if (typeof Symbol === "function" && typeof Symbol.iterator === "symbol") {
    _typeof = function (obj) {
      return typeof obj;
    };
  } else {
    _typeof = function (obj) {
      return obj && typeof Symbol === "function" && obj.constructor === Symbol && obj !== Symbol.prototype ? "symbol" : typeof obj;
    };
  }

  return _typeof(obj);
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

function _objectWithoutPropertiesLoose(source, excluded) {
  if (source == null) return {};
  var target = {};
  var sourceKeys = Object.keys(source);
  var key, i;

  for (i = 0; i < sourceKeys.length; i++) {
    key = sourceKeys[i];
    if (excluded.indexOf(key) >= 0) continue;
    target[key] = source[key];
  }

  return target;
}

function _objectWithoutProperties(source, excluded) {
  if (source == null) return {};

  var target = _objectWithoutPropertiesLoose(source, excluded);

  var key, i;

  if (Object.getOwnPropertySymbols) {
    var sourceSymbolKeys = Object.getOwnPropertySymbols(source);

    for (i = 0; i < sourceSymbolKeys.length; i++) {
      key = sourceSymbolKeys[i];
      if (excluded.indexOf(key) >= 0) continue;
      if (!Object.prototype.propertyIsEnumerable.call(source, key)) continue;
      target[key] = source[key];
    }
  }

  return target;
}

function _slicedToArray(arr, i) {
  return _arrayWithHoles(arr) || _iterableToArrayLimit(arr, i) || _unsupportedIterableToArray(arr, i) || _nonIterableRest();
}

function _toConsumableArray(arr) {
  return _arrayWithoutHoles(arr) || _iterableToArray(arr) || _unsupportedIterableToArray(arr) || _nonIterableSpread();
}

function _arrayWithoutHoles(arr) {
  if (Array.isArray(arr)) return _arrayLikeToArray(arr);
}

function _arrayWithHoles(arr) {
  if (Array.isArray(arr)) return arr;
}

function _iterableToArray(iter) {
  if (typeof Symbol !== "undefined" && iter[Symbol.iterator] != null || iter["@@iterator"] != null) return Array.from(iter);
}

function _iterableToArrayLimit(arr, i) {
  var _i = arr && (typeof Symbol !== "undefined" && arr[Symbol.iterator] || arr["@@iterator"]);

  if (_i == null) return;
  var _arr = [];
  var _n = true;
  var _d = false;

  var _s, _e;

  try {
    for (_i = _i.call(arr); !(_n = (_s = _i.next()).done); _n = true) {
      _arr.push(_s.value);

      if (i && _arr.length === i) break;
    }
  } catch (err) {
    _d = true;
    _e = err;
  } finally {
    try {
      if (!_n && _i["return"] != null) _i["return"]();
    } finally {
      if (_d) throw _e;
    }
  }

  return _arr;
}

function _unsupportedIterableToArray(o, minLen) {
  if (!o) return;
  if (typeof o === "string") return _arrayLikeToArray(o, minLen);
  var n = Object.prototype.toString.call(o).slice(8, -1);
  if (n === "Object" && o.constructor) n = o.constructor.name;
  if (n === "Map" || n === "Set") return Array.from(o);
  if (n === "Arguments" || /^(?:Ui|I)nt(?:8|16|32)(?:Clamped)?Array$/.test(n)) return _arrayLikeToArray(o, minLen);
}

function _arrayLikeToArray(arr, len) {
  if (len == null || len > arr.length) len = arr.length;

  for (var i = 0, arr2 = new Array(len); i < len; i++) arr2[i] = arr[i];

  return arr2;
}

function _nonIterableSpread() {
  throw new TypeError("Invalid attempt to spread non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

function _nonIterableRest() {
  throw new TypeError("Invalid attempt to destructure non-iterable instance.\nIn order to be iterable, non-array objects must have a [Symbol.iterator]() method.");
}

var local = {
  second: d3Time.timeSecond,
  minute: d3Time.timeMinute,
  hour: d3Time.timeHour,
  day: d3Time.timeDay,
  week: d3Time.timeWeek,
  month: d3Time.timeMonth,
  year: d3Time.timeYear
};
var utc = {
  second: d3Time.utcSecond,
  minute: d3Time.utcMinute,
  hour: d3Time.utcHour,
  day: d3Time.utcDay,
  week: d3Time.utcWeek,
  month: d3Time.utcMonth,
  year: d3Time.utcYear
};
var formats = {
  millisecond: function millisecond() {
    return '.%L';
  },
  second: function second() {
    return ':%S';
  },
  minute: function minute(_ref) {
    var use24h = _ref.use24h;
    return "%".concat(use24h ? 'H' : 'I', ":%M");
  },
  hour: function hour(_ref2) {
    var use24h = _ref2.use24h;
    return use24h ? '%H:%M' : '%I %p';
  },
  day: function day() {
    return '%a %d';
  },
  week: function week() {
    return '%b %d';
  },
  month: function month() {
    return '%b';
  },
  year: function year() {
    return '%Y';
  }
};
function axisTimeFormatter (_ref3) {
  var _ref3$useUtc = _ref3.useUtc,
      useUtc = _ref3$useUtc === void 0 ? false : _ref3$useUtc,
      _ref3$use24h = _ref3.use24h,
      use24h = _ref3$use24h === void 0 ? false : _ref3$use24h;
  var formatOpts = {
    use24h: use24h
  };
  var formatFn = useUtc ? d3TimeFormat.utcFormat : d3TimeFormat.timeFormat;
  var f = Object.assign.apply(Object, [{}].concat(_toConsumableArray(Object.entries(formats).map(function (_ref4) {
    var _ref5 = _slicedToArray(_ref4, 2),
        unit = _ref5[0],
        fn = _ref5[1];

    return _defineProperty({}, unit, formatFn(fn(formatOpts)));
  }))));
  var t = useUtc ? utc : local;
  return function (date) {
    return (t.second(date) < date ? f.millisecond : t.minute(date) < date ? f.second : t.hour(date) < date ? f.minute : t.day(date) < date ? f.hour : t.month(date) < date ? t.week(date) < date ? f.day : f.week : t.year(date) < date ? f.month : f.year)(date);
  };
}

var _excluded = ["x", "y"];
var AXIS_HEIGHT = 20;
var MAX_FONT_SIZE = 13;
var MIN_SERIES_HEIGHT_WITH_BORDER = 20;
var MAX_ZOOM_SCALE = 5000;
var horizonChart = Kapsule__default['default']({
  props: {
    width: {
      "default": window.innerWidth
    },
    height: {
      "default": window.innerHeight
    },
    data: {
      "default": []
    },
    ts: {
      "default": 'ts'
    },
    val: {
      "default": 'val'
    },
    series: {},
    seriesComparator: {
      "default": function _default(a, b) {
        return a.localeCompare(b);
      }
    },
    horizonBands: {
      "default": 4
    },
    horizonMode: {
      "default": 'offset'
    },
    // or mirror
    useUtc: {
      "default": false
    },
    // local timezone vs utc
    use24h: {
      "default": true
    },
    // 24h vs am/pm format
    yExtent: {},
    // undefined means it will be derived dynamically from the data
    yNormalize: {
      "default": false
    },
    yScaleExp: {
      "default": 1
    },
    yAggregation: {
      "default": function _default(vals) {
        return vals.reduce(function (agg, val) {
          return agg + val;
        });
      }
    },
    // sum reduce
    positiveColors: {
      "default": ['white', 'midnightblue']
    },
    negativeColors: {
      "default": ['white', 'crimson']
    },
    positiveColorStops: {},
    negativeColorStops: {},
    interpolationCurve: {
      "default": d3Shape.curveBasis
    },
    seriesLabelFormatter: {
      "default": function _default(series) {
        return series;
      }
    },
    showRuler: {
      "default": true,
      triggerUpdate: false,
      onChange: function onChange(show, state) {
        return state.ruler && state.ruler.style('visibility', show ? 'visible' : 'hidden');
      }
    },
    enableZoom: {
      "default": false
    },
    tooltipContent: {
      "default": function _default(_ref) {
        var series = _ref.series,
            ts = _ref.ts,
            val = _ref.val,
            useUtc = _ref.useUtc,
            use24h = _ref.use24h;
        return "<b>".concat(series, "</b><br>\n      ").concat(new Date(ts).toLocaleString(undefined, {
          timeZone: useUtc ? 'UTC' : undefined,
          hour12: !use24h
        }), ":\n      <b>").concat(val, "</b>");
      }
    },
    transitionDuration: {
      "default": 250
    },
    onHover: {
      triggerUpdate: false
    },
    onClick: {}
  },
  stateInit: function stateInit() {
    return {
      horizonLayouts: {},
      // per series
      zoomedInteraction: false,
      timeAxis: d3Axis.axisBottom()
    };
  },
  init: function init(el, state) {
    var isD3Selection = !!el && _typeof(el) === 'object' && !!el.node && typeof el.node === 'function';
    var d3El = d3Selection.select(isD3Selection ? el.node() : el);
    d3El.html(null); // wipe DOM

    state.chartsEl = d3El.append('div').style('position', 'relative');
    state.ruler = state.chartsEl.append('div').attr('class', 'horizon-ruler');
    state.axisEl = d3El.append('svg').attr('height', AXIS_HEIGHT).style('display', 'block'); // Bind zoom

    state.chartsEl.call(state.zoom = d3Zoom.zoom());
    state.zoom.__baseElem = state.chartsEl; // Attach controlling elem for easy access

    state.zoom.on('zoom', function () {
      if (state.enableZoom) {
        state.zoomedInteraction = true;

        state._rerender();
      }
    });
  },
  update: function update(state) {
    var valAccessor = accessorFn__default['default'](state.val);
    var yExtentAccessor = accessorFn__default['default'](state.yExtent);
    var yScaleExpAccessor = accessorFn__default['default'](state.yScaleExp);
    var positiveColorsAccessor = accessorFn__default['default'](state.positiveColors);
    var negativeColorsAccessor = accessorFn__default['default'](state.negativeColors);
    var positiveColorStopsAccessor = accessorFn__default['default'](state.positiveColorStops);
    var negativeColorStopsAccessor = accessorFn__default['default'](state.negativeColorStops); // memoize to prevent calling timeAccessor multiple times

    var tsMemo = memo__default['default'](accessorFn__default['default'](state.ts));
    var times = state.data.map(tsMemo);
    var timeScale = (state.useUtc ? d3Scale.scaleUtc : d3Scale.scaleTime)().domain(d3Array.extent(times)).range([0, state.width]); // const timeAxis = d3AxisBottom(timeScale)

    state.timeAxis.scale(timeScale).tickFormat(axisTimeFormatter({
      useUtc: state.useUtc,
      use24h: state.use24h
    })); // set scale extent to 1 if zoom is disabled, to allow default wheel events to bubble

    state.zoom.scaleExtent([1, state.enableZoom ? MAX_ZOOM_SCALE : 1]);

    if (state.enableZoom) {
      state.zoom.translateExtent([[0, 0], [state.width, 0]]); // apply zoom

      var zoomTransform = d3Zoom.zoomTransform(state.zoom.__baseElem.node());
      var zoomedTimeLength = (timeScale.domain()[1] - timeScale.domain()[0]) / zoomTransform.k;
      var zoomedStartTime = timeScale.invert(-zoomTransform.x / zoomTransform.k);
      timeScale.domain([zoomedStartTime, new Date(+zoomedStartTime + zoomedTimeLength)]);
    }

    var bySeries = indexBy__default['default'](state.data, state.series);
    var seriesData = Object.keys(bySeries).sort(state.seriesComparator).map(function (series) {
      return {
        series: series,
        data: bySeries[series]
      };
    }); // Calculate global Y max if yNormalize is on

    var normalizedYMax = !state.yNormalize ? null : d3Array.max(indexBy__default['default'](state.data, [state.series, function (d) {
      return +tsMemo(d);
    }], function (points) {
      return state.yAggregation(points.map(valAccessor));
    }, true).map(function (d) {
      return d.vals;
    }).map(Math.abs) // get max absolute val
    );
    var seriesHeight = (state.height - AXIS_HEIGHT) / seriesData.length;
    var fontSize = Math.min(MAX_FONT_SIZE, Math.round(seriesHeight * 0.9));
    var borderWidth = seriesHeight >= MIN_SERIES_HEIGHT_WITH_BORDER ? 1 : 0;
    var tr = d3Transition.transition().duration(state.zoomedInteraction ? 0 : state.transitionDuration);
    var horizons = state.chartsEl.selectAll('.horizon-series').data(seriesData, function (d) {
      return d.series;
    });
    horizons.exit().each(function (_ref2) {
      var series = _ref2.series;
      state.horizonLayouts[series].height(0);
      delete state.horizonLayouts[series];
    }).transition(tr).remove();
    var newHorizons = horizons.enter().append('div').attr('class', 'horizon-series');
    newHorizons.append('div').attr('class', 'chart').each(function (_ref3) {
      var series = _ref3.series;
      // instantiate d3-horizon for each new series
      state.horizonLayouts[series] = d3Horizon__default['default']().width(state.width).height(seriesHeight - borderWidth)(this);
    });
    newHorizons.append('span').attr('class', 'label').style('font-size', 0);
    var allHorizons = horizons.merge(newHorizons).order();
    allHorizons.transition(tr).style('width', state.width).style('border-top-width', "".concat(borderWidth, "px"));
    allHorizons.select('.chart').each(function (_ref4) {
      var series = _ref4.series,
          data = _ref4.data;
      return state.horizonLayouts[series].width(state.width).height(seriesHeight - borderWidth).data(data).bands(state.horizonBands).mode(state.horizonMode).x(tsMemo).y(valAccessor).yExtent(yExtentAccessor(series) || normalizedYMax || null) // Use normalizedYMax (if calculated) if yExtent is not explicitly set
      .yScaleExp(yScaleExpAccessor(series)).yAggregation(state.yAggregation).xMin(timeScale.domain()[0]).xMax(timeScale.domain()[1]).positiveColors(positiveColorsAccessor(series)).negativeColors(negativeColorsAccessor(series)).positiveColorStops(positiveColorStopsAccessor(series)).negativeColorStops(negativeColorStopsAccessor(series)).interpolationCurve(state.interpolationCurve).duration(state.transitionDuration).tooltipContent(state.tooltipContent && function (_ref5) {
        var x = _ref5.x,
            y = _ref5.y,
            rest = _objectWithoutProperties(_ref5, _excluded);

        return state.tooltipContent(_objectSpread2(_objectSpread2({
          series: series,
          ts: x,
          val: y
        }, rest), {}, {
          useUtc: state.useUtc,
          use24h: state.use24h
        }));
      }).onHover(function (d) {
        d && state.ruler.style('left', "".concat(timeScale(d.x), "px"));
        state.ruler.style('opacity', d ? 0.2 : 0);
        d && state.onHover && state.onHover({
          series: series,
          ts: d.x,
          val: d.y,
          points: d.points
        });
      }).onClick(state.onClick ? function (d) {
        return d && state.onClick({
          series: series,
          ts: d.x,
          val: d.y,
          points: d.points
        });
      } : undefined);
    });
    allHorizons.select('.label').transition(tr).style('font-size', "".concat(fontSize, "px")).text(function (d) {
      return state.seriesLabelFormatter(d.series);
    });
    state.axisEl.call(state.timeAxis).transition(tr).attr('width', state.width);
    state.zoomedInteraction = false;
  }
});

module.exports = horizonChart;
